use std::sync::atomic::{AtomicI64, Ordering};

use js_sys::Date;
use log::{error, trace};
use md5::{Digest, Md5};
use rsa::rand_core::OsRng;
use rsa::{pkcs8::DecodePublicKey, Pkcs1v15Encrypt, RsaPublicKey};
use serde::{Deserialize, Serialize};
use wasm_bindgen::prelude::*;
use wasm_bindgen_futures::JsFuture;
use web_sys::{Request, RequestInit, RequestMode, Response, Window};

const PUB_PEM: &str = "-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCSvADFwZL6joQNvy9A66nE7Hf4
RbWj7YWPmENyvzk4pazchwQkIsVdCAXkRIrFlZ9cy8426IKOgRDAmfWM2owgz0fL
5YryigTwHgEn2y0VfCdY/8gnvYYdgliBJdhNCZC1d/ZsF4hRzIXR09cVGqe2wW40
4lEYsZNCFIjbmM7y6QIDAQAB
-----END PUBLIC KEY-----";

const COMMON_UA: &str = "9e2ee1d8183042a1";

static SERVER_TIME: AtomicI64 = AtomicI64::new(0);
static LOCAL_TIME: AtomicI64 = AtomicI64::new(0);

#[derive(Serialize, Deserialize, Debug)]
pub struct UnifyData {
    /// 接口地址
    #[serde(default = "default_vp")]
    pub vp: String,
    /// 方法类型
    #[serde(default = "default_vm")]
    pub vm: String,
    /// userAgent
    pub ua: Option<String>,
    /// 用户token[未使用]
    #[serde(default = "default_vc")]
    pub vc: String,
    /// 用户传入时间[未使用]
    pub vt: Option<i32>,
    /// [未使用]
    pub uc: Option<String>,
}

fn default_vc() -> String {
    "no_token".to_string()
}

fn default_vp() -> String {
    "none".to_string()
}

fn default_vm() -> String {
    "none".to_string()
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Unify {
    pub p: String,
    pub t: i64,
    pub u: String,
}

#[wasm_bindgen]
pub fn x2c8dn(data: i64) {
    trace!("initial_sever_time: {}", data);
    SERVER_TIME.swap(data, Ordering::SeqCst);
}

/// need browser env
#[wasm_bindgen]
pub async fn q2c8dn(vf: JsValue) -> Result<String, JsValue> {
    let unify_data = serde_wasm_bindgen::from_value::<UnifyData>(vf);
    trace!("{:?}", unify_data);
    if let Ok(ud) = unify_data {
        let window = web_sys::window();
        let sign = match window {
            Some(win) => {
                check_server_time(&win).await;
                encrypt_api(ud, &win)
            }
            None => Err(JsValue::from_str("window not found")),
        };
        if let Ok(sign) = sign {
            return Ok(sign);
        }
        error!("encrypt api failed: {:?}", sign)
    }
    Ok(String::from(""))
}
/// need browser env
#[wasm_bindgen]
pub async fn ixq2pl() -> Result<(), JsValue> {
    match web_sys::window() {
        Some(win) => {
            check_server_time(&win).await;
            Ok(())
        }
        None => {
            error!("{}", "window not found");
            Err(JsValue::from_str("window not found"))
        }
    }
}

/// need browser or node env
#[wasm_bindgen]
pub fn eligo(vp: &str, vt: &str) -> String {
    trace!("eligo: [vp: {}, vt: {}]", vp, vt);
    let path_md5 = get_md5_middle16(&vp);
    let sign = match web_sys::window() {
        Some(win) => match get_ua(&win) {
            Ok(ua) => cal_sign(path_md5, vt.to_string(), get_md5_middle16(&ua)),
            Err(_) => "".to_string(),
        },
        None => {
            cal_sign(path_md5, vt.to_string(), COMMON_UA.to_string())
        }
    };
    sign
}

async fn check_server_time(win: &Window) {
    if SERVER_TIME.load(Ordering::SeqCst) > 0 {
        return;
    }
    if let Ok(url) = get_url(win) {
        let data = get_server_time(&url, win).await;
        if let Ok(data) = data {
            SERVER_TIME.swap(data, Ordering::SeqCst);
            LOCAL_TIME.swap(Date::now() as i64, Ordering::SeqCst);
            return;
        }
        error!("{:?}", data);
    }
}

fn encrypt_api(unify_data: UnifyData, win: &Window) -> Result<String, JsValue> {
    let ua = get_ua(&win)?;
    let ua_md5 = get_md5_middle16(&ua);
    let path_md5 = get_md5_middle16(&unify_data.vp);
    let now = get_now_time();
    let unify = Unify {
        p: path_md5,
        t: now,
        u: ua_md5,
    };
    let a = encrypt_unify(unify);
    Ok(a)
}

fn encrypt_unify(data: Unify) -> String {
    cal_sign(data.p, data.t.to_string(), data.u)
}

fn cal_sign(p: String, t: String, u: String) -> String {
    let unify_str = format!("{}#{}#{}", p, t, u);
    let pub_key = RsaPublicKey::from_public_key_pem(PUB_PEM).unwrap();
    let enc_data = pub_key
        .encrypt(&mut OsRng, Pkcs1v15Encrypt, &unify_str.as_bytes())
        .expect("failed to encrypt");
    let sign = crate::utils::base64_encode(&enc_data);
    sign
}

fn get_md5_middle16(data: &str) -> String {
    let mut hasher = Md5::new();
    hasher.update(data.as_bytes());
    let result = format!("{:x}", hasher.finalize());
    let result = &result[8..24];
    result.to_string()
}

fn get_ua(window: &Window) -> Result<String, JsValue> {
    let navigator = js_sys::Reflect::get(window, &JsValue::from_str("navigator"))?;
    let ua = js_sys::Reflect::get(&navigator, &JsValue::from_str("userAgent"))?
        .as_string()
        .unwrap_or(String::from("no_user_agent"));
    Ok(ua)
}

fn get_now_time() -> i64 {
    let local_now = Date::now() as i64;
    let initial_local_time = LOCAL_TIME.load(Ordering::SeqCst);
    let initial_sever_time = SERVER_TIME.load(Ordering::SeqCst);
    trace!(
        "initial_local_time: {}, initial_sever_time: {}",
        initial_local_time,
        initial_sever_time
    );
    local_now - initial_local_time + initial_sever_time
}

fn get_url(window: &Window) -> Result<String, JsValue> {
    let location = js_sys::Reflect::get(window, &JsValue::from_str("location"))?;
    let host = js_sys::Reflect::get(&location, &JsValue::from_str("host"))?
        .as_string()
        .expect("host should exist");

    let protocol = js_sys::Reflect::get(&location, &JsValue::from_str("protocol"))?
        .as_string()
        .unwrap_or(String::from("http"));
    let now = Date::new_0().value_of().to_string();
    Ok(format!(
        "{}//{}{}?_s={}",
        protocol, host, "/api/admin/common/sys/time", &now
    ))
}

#[derive(Serialize, Deserialize, Debug)]
struct TimeData {
    pub success: bool,
    pub code: usize,
    pub result: Option<String>,
    pub timestamp: i64,
    #[serde(rename(serialize = "traceId", deserialize = "traceId"))]
    pub trace_id: Option<String>,
    pub data: Option<i64>,
}

async fn get_server_time(url: &str, window: &Window) -> Result<i64, JsValue> {
    let mut opts = RequestInit::new();
    opts.method("GET");
    opts.mode(RequestMode::Cors);
    let request = Request::new_with_str_and_init(&url, &opts)?;
    let resp_value = JsFuture::from(window.fetch_with_request(&request)).await?;

    let resp: Response = resp_value.dyn_into()?;
    if resp.status() == 200 {
        let data = JsFuture::from(resp.json()?).await?;
        let time_data = serde_wasm_bindgen::from_value::<TimeData>(data)?;
        trace!("{:?}", time_data);
        if !time_data.success {
            return Err(JsValue::from_str(
                &time_data.result.unwrap_or(String::from("invalid data")),
            ));
        }
        let time = time_data.data;
        if time.is_none() {
            return Ok(time_data.timestamp);
        } else {
            return Ok(time_data.data.unwrap());
        };
    }
    let err_msg = format!("{} {}", resp.status(), resp.status_text());
    Err(JsValue::from_str(&err_msg))
}
